# This is a class to manage variable environments


class Environment:
    def __init__(self, content=None, namespace=None):
        self._variables = {}

        if content is not None:
            self.update(content, namespace)

    def split_namespace(self, key):
        if "." not in key:
            return [key]

        return key.split(".")

    def setvar(self, key, value):
        path = self.split_namespace(key)

        if len(path) == 1:
            self._variables[path[0]] = value
            return

        namespace = self._variables
        for i in path[0:-1]:
            if i not in namespace:
                namespace[i] = {}

            namespace = namespace[i]

        namespace[path[-1]] = value

    def getvar(self, key, default=None):
        path = self.split_namespace(key)

        try:
            if len(path) == 1:
                return self._variables[path[0]]

            namespace = self._variables
            for i in path[0:-1]:
                namespace = namespace[i]

            return namespace[path[-1]]
        except KeyError:
            if default is None:
                raise

            return default

    def update(self, adict, namespace=None):
        if namespace is None:
            self._variables = adict
        else:
            self._variables[namespace] = adict

    def asdict(self):
        return self._variables



import argparse
import logging
import sys
from importlib import metadata

import yaml
from mau import ConfigurationError, Mau, load_visitors
from mau.lexers.base_lexer import LexerError
from mau.parsers.base_parser import ParserError
from mau.visitors.base_visitor import BaseVisitor, VisitorError
from mau.text_buffer.context import print_context
from tabulate import tabulate

__version__ = metadata.version("mau")
_logger = logging.getLogger(__name__)

visitor_classes = load_visitors()
visitors = {i.format_code: i for i in visitor_classes}


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input-file",
        action="store",
        required=True,
        help="Input file",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        action="store",
        required=False,
        help="Optional output file ('-' for standard output)",
    )

    parser.add_argument(
        "-c",
        "--config-file",
        action="store",
        required=False,
        help="Optional YAML config file",
    )

    parser.add_argument(
        "-f",
        "--format",
        action="store",
        required=True,
        choices=visitors.keys(),
        help="Output format",
    )

    parser.add_argument(
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )

    parser.add_argument(
        "--debug",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )

    parser.add_argument(
        "--lex-only",
        dest="lexonly",
        help="performs only lexing and prints tokens",
        action="store_true",
    )

    parser.add_argument(
        "--parse-only",
        dest="parseonly",
        help="performs only parsing and prints nodes",
        action="store_true",
    )

    parser.add_argument(
        "--version", action="version", version=f"Mau version {__version__}"
    )

    return parser.parse_args()


def setup_logging(loglevel):
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def write_output(output, output_file, transform=None):
    if transform:
        output = transform(output)

    # The output file can be "-" which means
    # the standard output
    if output_file == "-":
        print(output)

        return

    # We need to write on an actual file
    with open(output_file, "w", encoding="utf-8") as out:
        out.write(output)
        out.write("\n")


def main():
    # Get arguments and logging set up
    args = parse_args()
    setup_logging(args.loglevel)

    # Read the input file
    with open(args.input_file, "r", encoding="utf-8") as input_file:
        text = input_file.read()

    # Load the YAML configuration file into a dictionary
    config = {}
    if args.config_file:
        with open(args.config_file, "r", encoding="utf-8") as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)

    # The directory that contains custom templates.
    templates_directory = config.get("templates_directory", None)

    # Select the visitor class according to the target
    # format specified on the command line
    visitor_class = visitors[args.format]

    # This extracts the custom templates defined
    # in the config file
    custom_templates = config.get("custom_templates", None)

    # Find out the name of the output file
    output_file = args.output_file or args.input_file.replace(
        ".mau", visitor_class.extension
    )

    # The Mau object configured with what we figured out above.
    mau = Mau(
        args.input_file,
        visitor_class=visitor_class,
        config=config,
        custom_templates=custom_templates,
        templates_directory=templates_directory,
        full_document=True,
    )

    # Run the lexer on the input data
    try:
        lexer = mau.run_lexer(text)
    except LexerError as exception:
        print_context(exception.context, f"Lexer error - {str(exception)}")
        sys.exit(1)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        print(f"Unhandled exception: {exception}")
        sys.exit(1)

    if args.lexonly:
        write_output(
            tabulate(
                [(t.type, t.value, t.context) for t in lexer.tokens],
                maxcolwidths=[10, 60, 30],
            ),
            output_file,
        )

        sys.exit(1)

    # Run the parser on the tokens
    try:
        parser = mau.run_parser(lexer.tokens)
    except ParserError as exception:
        print_context(exception.token.context, f"Parser error - {str(exception)}")
        sys.exit(1)

    if args.parseonly:
        write_output(parser.nodes, output_file)
        sys.exit(1)

    # Run the visitor on the AST
    try:
        output = mau.process(parser.nodes, parser.environment)
    except ConfigurationError as exception:
        print(f"Configuration error: {exception}")
        sys.exit(1)
    except ParserError as exception:
        print_context(exception.token.context, f"Parser error - {str(exception)}")
        sys.exit(1)
    except VisitorError as exception:
        print(exception)
        sys.exit(1)

    write_output(output, output_file, transform=visitor_class.transform)



from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, StyleNode, TextNode
from mau.nodes.page import BlockNode, ParagraphNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_admonition():
    source = """
    [admonition,someclass,someicon,somelabel]
    ----
    Content
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="admonition",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content"),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"class": "someclass", "icon": "someicon", "label": "somelabel"},
        )
    ]


def test_parse_block_quote_attribution_in_secondary_content():
    source = """
    [quote]
    ----
    Learn about the Force, Luke.
    ----
    _Star Wars_, 1977
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="quote",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Learn about the Force, Luke."),
                        ]
                    )
                ),
            ],
            secondary_content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            StyleNode(
                                "underscore",
                                SentenceNode(
                                    [
                                        TextNode("Star Wars"),
                                    ]
                                ),
                            ),
                            TextNode(", 1977"),
                        ]
                    )
                ),
            ],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]



from mau.nodes.footnotes import CommandFootnotesNode, FootnotesEntryNode
from mau.nodes.inline import ListItemNode, SentenceNode, TextNode
from mau.nodes.page import (
    BlockNode,
    CommandTocNode,
    ContentImageNode,
    ContentNode,
    DocumentNode,
    HeaderNode,
    HorizontalRuleNode,
    ListNode,
    ParagraphNode,
)
from mau.nodes.references import CommandReferencesNode, ReferencesEntryNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode
from mau.visitors.base_visitor import BaseVisitor


def test_horizontal_rule_node():
    visitor = BaseVisitor()

    node = HorizontalRuleNode(
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "horizontal_rule",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_paragraph_node():
    visitor = BaseVisitor()

    node = ParagraphNode(
        TextNode("Just some text"),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "paragraph",
            "content": {
                "data": {
                    "type": "text",
                    "value": "Just some text",
                }
            },
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_header_node():
    visitor = BaseVisitor()

    node = HeaderNode(
        value="Just some text",
        level="3",
        anchor="someanchor",
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "header",
            "value": "Just some text",
            "level": 3,
            "anchor": "someanchor",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_list_node():
    visitor = BaseVisitor()

    node = ListNode(
        ordered=True,
        items=[ListItemNode("4", TextNode("Just some text."))],
        main_node=True,
        args=["arg1", "arg2"],
        kwargs={"key1": "value1", "start": 4},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "list",
            "items": [
                {
                    "data": {
                        "type": "list_item",
                        "level": 4,
                        "content": {
                            "data": {
                                "type": "text",
                                "value": "Just some text.",
                            }
                        },
                    }
                }
            ],
            "main_node": True,
            "ordered": True,
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1", "start": 4},
            "tags": ["tag1", "tag2"],
        }
    }


def test_content_node():
    visitor = BaseVisitor()

    node = ContentNode(
        content_type="sometype",
        title=TextNode("sometitle"),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "content",
            "content_type": "sometype",
            "title": {
                "data": {
                    "type": "text",
                    "value": "sometitle",
                }
            },
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_content_image_node():
    visitor = BaseVisitor()

    node = ContentImageNode(
        uri="someuri",
        alt_text="sometext",
        classes=["class1", "class2"],
        title=TextNode("sometitle"),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "content_image",
            "uri": "someuri",
            "title": {
                "data": {
                    "type": "text",
                    "value": "sometitle",
                }
            },
            "alt_text": "sometext",
            "classes": ["class1", "class2"],
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_block_node_standard_block_template():
    visitor = BaseVisitor()

    node = BlockNode(
        blocktype="someblock",
        content=[TextNode("my content")],
        secondary_content=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=TextNode("sometitle"),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "block",
            "blocktype": "someblock",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "my content",
                    }
                }
            ],
            "secondary_content": [
                {
                    "data": {
                        "type": "text",
                        "value": "my secondary content",
                    }
                }
            ],
            "classes": ["class1", "class2"],
            "title": {
                "data": {
                    "type": "text",
                    "value": "sometitle",
                }
            },
            "engine": "someengine",
            "preprocessor": "somepreprocessor",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_command_toc_node():
    visitor = BaseVisitor()

    node = CommandTocNode(
        entries=[
            HeaderNode("Header 1", "1", "header-1"),
            HeaderNode("Header 1.1", "2", "header-1-1"),
            HeaderNode("Header 2", "1", "header-2"),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "command_toc",
            "entries": [
                {
                    "data": {
                        "anchor": "header-1",
                        "args": ["arg1", "arg2"],
                        "children": [
                            {
                                "data": {
                                    "anchor": "header-1-1",
                                    "args": ["arg1", "arg2"],
                                    "children": [],
                                    "kwargs": {"key1": "value1"},
                                    "tags": ["tag1", "tag2"],
                                    "type": "toc_entry",
                                    "value": "Header 1.1",
                                }
                            },
                        ],
                        "kwargs": {"key1": "value1"},
                        "tags": ["tag1", "tag2"],
                        "type": "toc_entry",
                        "value": "Header 1",
                    }
                },
                {
                    "data": {
                        "type": "toc_entry",
                        "anchor": "header-2",
                        "args": ["arg1", "arg2"],
                        "children": [],
                        "kwargs": {"key1": "value1"},
                        "tags": ["tag1", "tag2"],
                        "value": "Header 2",
                    }
                },
            ],
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_command_toc_node_exclude():
    visitor = BaseVisitor()

    node = CommandTocNode(
        entries=[
            HeaderNode("Header 1", "1", "header-1", tags=["notoc"]),
            HeaderNode("Header 1.1", "2", "header-1-1"),
            HeaderNode("Header 2", "1", "header-2"),
        ],
        args=["arg1", "arg2"],
        kwargs={"exclude_tag": "notoc"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "command_toc",
            "entries": [
                {
                    "data": {
                        "type": "toc_entry",
                        "anchor": "header-2",
                        "args": ["arg1", "arg2"],
                        "children": [],
                        "kwargs": {"exclude_tag": "notoc"},
                        "tags": ["tag1", "tag2"],
                        "value": "Header 2",
                    }
                },
            ],
            "args": ["arg1", "arg2"],
            "kwargs": {"exclude_tag": "notoc"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_command_footnotes_node():
    visitor = BaseVisitor()

    node = CommandFootnotesNode(
        entries=[
            FootnotesEntryNode(
                [TextNode("Footnote 1")], "1", "anchor-1", "anchor-1-def"
            ),
            FootnotesEntryNode(
                [TextNode("Footnote 2")], "2", "anchor-2", "anchor-2-def"
            ),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "command_footnotes",
            "entries": [
                {
                    "data": {
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Footnote 1",
                                }
                            },
                        ],
                        "content_anchor": "anchor-1-def",
                        "number": "1",
                        "reference_anchor": "anchor-1",
                        "type": "footnotes_entry",
                    }
                },
                {
                    "data": {
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Footnote 2",
                                }
                            },
                        ],
                        "content_anchor": "anchor-2-def",
                        "number": "2",
                        "reference_anchor": "anchor-2",
                        "type": "footnotes_entry",
                    }
                },
            ],
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_command_references():
    visitor = BaseVisitor()

    node = CommandReferencesNode(
        content_type="content_type1",
        name="value1",
        category=None,
        entries={
            ("content_type1", "value1"): ReferencesEntryNode(
                "content_type1",
                "value1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                reference_anchor="ref-content_type1-1-XXYY",
                content_anchor="cnt-content_type1-1-XXYY",
            ),
            ("content_type1", "value2"): ReferencesEntryNode(
                "content_type1",
                "value2",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 2"),
                            ]
                        )
                    ),
                ],
                number=2,
                reference_anchor="ref-content_type1-2-XXYY",
                content_anchor="cnt-content_type1-2-XXYY",
            ),
            ("content_type2", "value1"): ReferencesEntryNode(
                "content_type2",
                "value1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 2, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                reference_anchor="ref-content_type2-1-XXYY",
                content_anchor="cnt-content_type2-1-XXYY",
            ),
        },
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "args": ["arg1", "arg2"],
            "content_type": "content_type1",
            "category": None,
            "entries": [
                {
                    "data": {
                        "category": None,
                        "content": [
                            {
                                "data": {
                                    "args": [],
                                    "content": {
                                        "data": {
                                            "content": [
                                                {
                                                    "data": {
                                                        "type": "text",
                                                        "value": "Content type 1, value 1",
                                                    }
                                                },
                                            ],
                                            "type": "sentence",
                                        }
                                    },
                                    "kwargs": {},
                                    "tags": [],
                                    "type": "paragraph",
                                }
                            },
                        ],
                        "content_anchor": "cnt-content_type1-1-XXYY",
                        "content_type": "content_type1",
                        "name": "value1",
                        "number": 1,
                        "reference_anchor": "ref-content_type1-1-XXYY",
                        "title": {},
                        "type": "references_entry",
                    }
                },
            ],
            "kwargs": {"key1": "value1"},
            "name": "value1",
            "tags": ["tag1", "tag2"],
            "type": "command_references",
        }
    }


def test_command_references_category():
    visitor = BaseVisitor()

    node = CommandReferencesNode(
        content_type="content_type1",
        name=None,
        category="cat1",
        entries={
            ("content_type1", "value1"): ReferencesEntryNode(
                "content_type1",
                "value1",
                category="cat1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                reference_anchor="ref-content_type1-1-XXYY",
                content_anchor="cnt-content_type1-1-XXYY",
            ),
            ("content_type1", "value2"): ReferencesEntryNode(
                "content_type1",
                "value2",
                category="cat2",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 2"),
                            ]
                        )
                    ),
                ],
                number=2,
                reference_anchor="ref-content_type1-2-XXYY",
                content_anchor="cnt-content_type1-2-XXYY",
            ),
            ("content_type1", "value3"): ReferencesEntryNode(
                "content_type1",
                "value3",
                category="cat1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 3"),
                            ]
                        )
                    ),
                ],
                number=1,
                reference_anchor="ref-content_type1-3-XXYY",
                content_anchor="cnt-content_type1-3-XXYY",
            ),
        },
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "args": ["arg1", "arg2"],
            "content_type": "content_type1",
            "category": "cat1",
            "entries": [
                {
                    "data": {
                        "category": "cat1",
                        "content": [
                            {
                                "data": {
                                    "args": [],
                                    "content": {
                                        "data": {
                                            "content": [
                                                {
                                                    "data": {
                                                        "type": "text",
                                                        "value": "Content type 1, value 1",
                                                    }
                                                },
                                            ],
                                            "type": "sentence",
                                        }
                                    },
                                    "kwargs": {},
                                    "tags": [],
                                    "type": "paragraph",
                                }
                            },
                        ],
                        "content_anchor": "cnt-content_type1-1-XXYY",
                        "content_type": "content_type1",
                        "name": "value1",
                        "number": 1,
                        "reference_anchor": "ref-content_type1-1-XXYY",
                        "title": {},
                        "type": "references_entry",
                    }
                },
                {
                    "data": {
                        "category": "cat1",
                        "content": [
                            {
                                "data": {
                                    "args": [],
                                    "content": {
                                        "data": {
                                            "content": [
                                                {
                                                    "data": {
                                                        "type": "text",
                                                        "value": "Content type 1, value 3",
                                                    }
                                                },
                                            ],
                                            "type": "sentence",
                                        }
                                    },
                                    "kwargs": {},
                                    "tags": [],
                                    "type": "paragraph",
                                }
                            },
                        ],
                        "content_anchor": "cnt-content_type1-3-XXYY",
                        "content_type": "content_type1",
                        "name": "value3",
                        "number": 1,
                        "reference_anchor": "ref-content_type1-3-XXYY",
                        "title": {},
                        "type": "references_entry",
                    }
                },
            ],
            "kwargs": {"key1": "value1"},
            "name": None,
            "tags": ["tag1", "tag2"],
            "type": "command_references",
        }
    }


def test_source_node():
    visitor = BaseVisitor()

    node = SourceNode(
        language="somelang",
        code=[
            TextNode("import sys"),
            TextNode("import: os"),
            TextNode(""),
            TextNode('print(os.environ["HOME"])'),
        ],
        markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
        callouts=[
            CalloutsEntryNode("imp", "This is an import"),
            CalloutsEntryNode("env", "Environment variables are paramount"),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "source",
            "blocktype": "default",
            "callouts": [
                {
                    "data": {
                        "marker": "imp",
                        "type": "callouts_entry",
                        "value": "This is an import",
                    }
                },
                {
                    "data": {
                        "marker": "env",
                        "type": "callouts_entry",
                        "value": "Environment variables are paramount",
                    }
                },
            ],
            "classes": [],
            "code": [
                {"data": {"type": "text", "value": "import sys"}},
                {"data": {"type": "text", "value": "import: os"}},
                {"data": {"type": "text", "value": ""}},
                {
                    "data": {
                        "type": "text",
                        "value": 'print(os.environ["HOME"])',
                    }
                },
            ],
            "highlights": [],
            "language": "somelang",
            "lines": 4,
            "markers": [
                None,
                {"data": {"type": "callout", "line": 1, "marker": "imp"}},
                None,
                {"data": {"type": "callout", "line": 3, "marker": "env"}},
            ],
            "preprocessor": None,
            "title": {},
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_document_node():
    visitor = BaseVisitor()

    node = DocumentNode(
        content=[
            ParagraphNode(
                TextNode("Just some text"),
                args=["arg1", "arg2"],
                kwargs={"key1": "value1"},
                tags=["tag1", "tag2"],
            )
        ],
        args=["arg3", "arg4"],
        kwargs={"key2": "value2"},
        tags=["tag3", "tag4"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "document",
            "content": [
                {
                    "data": {
                        "type": "paragraph",
                        "content": {
                            "data": {
                                "type": "text",
                                "value": "Just some text",
                            }
                        },
                        "args": ["arg1", "arg2"],
                        "kwargs": {"key1": "value1"},
                        "tags": ["tag1", "tag2"],
                    }
                }
            ],
            "args": ["arg3", "arg4"],
            "kwargs": {"key2": "value2"},
            "tags": ["tag3", "tag4"],
        }
    }



import textwrap

from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer


def dedent(text):
    return textwrap.dedent(text).strip()


def init_parser_factory(lexer_class, parser_class):
    """
    A factory that returns a parser initialiser.
    The returned function initialises and runs the lexer,
    initialises the parser and returns it.
    """

    def _init_parser(text, *args, **kwargs):
        text_buffer = TextBuffer(text, Context(source="main"))

        lex = lexer_class(text_buffer)
        lex.process()

        par = parser_class(lex.tokens, *args, **kwargs)

        return par

    return _init_parser


def parser_runner_factory(lexer_class, parser_class, *args, **kwds):
    """
    A factory that returns a parser runner.
    The returned function runs the parser on the given source.
    """

    init_parser = init_parser_factory(lexer_class, parser_class)

    def _run(source, **kwargs):
        kwds.update(kwargs)

        parser = init_parser(textwrap.dedent(source), *args, **kwds)
        parser.parse()

        return parser

    return _run



from mau.nodes.inline import (
    ClassNode,
    ImageNode,
    LinkNode,
    ListItemNode,
    MacroNode,
    SentenceNode,
    StyleNode,
    TextNode,
    VerbatimNode,
    WordNode,
)


def test_word_node():
    node = WordNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "word"
    assert node == WordNode("somevalue")


def test_text_node():
    node = TextNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "text"
    assert node == TextNode("somevalue")


def test_verbatim_node():
    node = VerbatimNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "verbatim"
    assert node == VerbatimNode("somevalue")


def test_sentence_node():
    node = SentenceNode([VerbatimNode("somevalue"), TextNode("othervalue")])

    assert node.content == [VerbatimNode("somevalue"), TextNode("othervalue")]
    assert node.node_type == "sentence"
    assert node == SentenceNode([VerbatimNode("somevalue"), TextNode("othervalue")])


def test_style_node():
    node = StyleNode("mystyle", TextNode("othervalue"))

    assert node.value == "mystyle"
    assert node.content == TextNode("othervalue")
    assert node.node_type == "style"
    assert node == StyleNode("mystyle", TextNode("othervalue"))


def test_macro_node():
    node = MacroNode("aname", args=["value"], kwargs={"akey": "avalue"})

    assert node.name == "aname"
    assert node.args == ["value"]
    assert node.kwargs == {"akey": "avalue"}
    assert node.node_type == "macro"
    assert node == MacroNode("aname", args=["value"], kwargs={"akey": "avalue"})


def test_class_node():
    node = ClassNode(["class1", "class2"], TextNode("othervalue"))

    assert node.classes == ["class1", "class2"]
    assert node.content == TextNode("othervalue")
    assert node.node_type == "class"
    assert node == ClassNode(["class1", "class2"], TextNode("othervalue"))


def test_link_node():
    node = LinkNode("atarget", "sometext")

    assert node.target == "atarget"
    assert node.text == "sometext"
    assert node.node_type == "link"
    assert node == LinkNode("atarget", "sometext")


def test_link_node_no_text():
    node = LinkNode("atarget")

    assert node.target == "atarget"
    assert node.text is None
    assert node.node_type == "link"
    assert node == LinkNode("atarget")


def test_image_node():
    node = ImageNode("someuri", "somealttext", "width", "height")

    assert node.uri == "someuri"
    assert node.alt_text == "somealttext"
    assert node.width == "width"
    assert node.height == "height"
    assert node.node_type == "image"
    assert node == ImageNode("someuri", "somealttext", "width", "height")


def test_list_item_node():
    node = ListItemNode("3", "somecontent")

    assert node.level == "3"
    assert node.content == "somecontent"
    assert node.node_type == "list_item"
    assert node == ListItemNode("3", "somecontent")



from unittest.mock import patch

import pytest
from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import BlockNode, HeaderNode, ParagraphNode
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_attributes_block():
    source = """
    [blocktype,myattr1=value1]
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode(
                                "This is a simple line of text followed by another line of text"
                            ),
                        ]
                    )
                ),
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("And this is another paragraph"),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"myattr1": "value1"},
        )
    ]


def test_attributes_can_contain_variables():
    source = """
    :value:42

    [blocktype,myattr1={value}]
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode(
                                "This is a simple line of text followed by another line of text"
                            ),
                        ]
                    )
                ),
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("And this is another paragraph"),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"myattr1": "42"},
        )
    ]


def test_parse_block_title_and_attributes():
    source = """
    .Just a title
    [blocktype, name1=value1, name2=value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=SentenceNode(
                [
                    TextNode("Just a title"),
                ]
            ),
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"name1": "value1", "name2": "value2"},
        )
    ]


def test_parse_block_title_and_attributes_are_reset():
    source = """
    .Just a title
    [blocktype1, name1=value1, name2=value2]
    ----
    ----

    [blocktype2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype1",
            content=[],
            secondary_content=[],
            classes=[],
            title=SentenceNode(
                [
                    TextNode("Just a title"),
                ]
            ),
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"name1": "value1", "name2": "value2"},
        ),
        BlockNode(
            blocktype="blocktype2",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]


def test_block_classes_single_class():
    source = """
    [blocktype,classes=cls1]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=["cls1"],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]


def test_block_classes_multiple_classes():
    source = """
    [blocktype,classes="cls1,cls2"]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=["cls1", "cls2"],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]


def test_block_engine():
    source = """
    [blocktype,engine=someengine]
    ----
    ----
    """

    with pytest.raises(ParserError):
        runner(source)


def test_block_mau_has_no_external_variables():
    source = """
    :answer:42
    [block, engine=mau]
    ----
    The answer is {answer}.
    ----
    """

    with pytest.raises(ParserError):
        assert runner(source)


def test_block_raw_engine():
    source = """
    [block, engine=raw]
    ----
    Raw content
    on multiple lines
    ----
    Secondary content
    on multiple lines as well
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="block",
            content=[
                TextNode("Raw content"),
                TextNode("on multiple lines"),
            ],
            secondary_content=[
                TextNode("Secondary content"),
                TextNode("on multiple lines as well"),
            ],
            classes=[],
            title=None,
            engine="raw",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_block_default_engine_adds_headers_to_global_toc(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Global header

    [someblock]
    ----
    = Block header
    ----
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode("Global header", "1", "XXYY"),
        BlockNode(
            blocktype="someblock",
            content=[
                HeaderNode("Block header", "1", "XXYY"),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]

    assert par.headers == [
        HeaderNode("Global header", "1", "XXYY"),
        HeaderNode("Block header", "1", "XXYY"),
    ]


def test_block_positive_condition_matches():
    source = """
    :render:yes

    [block, condition="if:render:yes"]
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="block",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is a paragraph."),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_block_positive_condition_doesnt_match():
    source = """
    :render:no

    [block, condition="if:render:yes"]
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == []


def test_block_negative_condition_matches():
    source = """
    :render:yes

    [block, condition="ifnot:render:no"]
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="block",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is a paragraph."),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_block_negative_condition_doesnt_match():
    source = """
    :render:no

    [block, condition="ifnot:render:no"]
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == []


def test_block_condition_accepts_booleans():
    source = """
    :render:

    [block, condition="ifnot:render:"]
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == []


def test_block_condition_raises_exception():
    source = """
    [block, condition="if:render"]
    ----
    This is a paragraph.
    ----
    """

    with pytest.raises(ParserError):
        runner(source)


def test_block_condition_can_use_variable_namespace():
    source = """
    :flags.render:yes

    [block, condition="if:flags.render:yes"]
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="block",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is a paragraph."),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_command_defblock():
    source = """
    ::defblock:alias, myblock, language=python, engine=source
    """

    par = runner(source)

    assert par.block_aliases["alias"] == "myblock"
    assert par.block_defaults["alias"] == {"language": "python", "engine": "source"}


def test_command_defblock_source():
    source = """
    """

    par = runner(source)

    assert par.block_aliases["source"] == "default"
    assert par.block_defaults["source"] == {"language": "text", "engine": "source"}


def test_command_defblock_source_can_be_overridden():
    source = """
    ::defblock:source, source, language=python, engine=source
    """

    par = runner(source)

    assert par.block_aliases["source"] == "source"
    assert par.block_defaults["source"] == {"language": "python", "engine": "source"}


def test_command_defblock_no_args():
    source = """
    ::defblock:
    """

    with pytest.raises(ParserError):
        runner(source)


def test_command_defblock_single_arg():
    source = """
    ::defblock:source
    """

    with pytest.raises(ParserError):
        runner(source)


def test_block_definitions_are_used():
    source = """
    ::defblock:alias, blocktype, name1=value1, name2=value2

    [alias]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"name1": "value1", "name2": "value2"},
        ),
    ]


def test_block_definitions_local_kwargs_overwrite_defined_ones():
    source = """
    ::defblock:alias, blocktype, name1=value1, name2=value2

    [alias, name1=value99]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"name1": "value99", "name2": "value2"},
        ),
    ]


def test_block_definitions_local_args_are_used():
    # This is testing that local args are used even when
    # an alias is used. However, the test should fail as
    # blocks cannot have unnamed arguments.

    source = """
    ::defblock:alias, blocktype, name1=value1, name2=value2

    [alias, attr1, name1=value99]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=["attr1"],
            kwargs={"name1": "value99", "name2": "value2"},
        ),
    ]


def test_block_definitions_unnamed_args_are_used_as_names():
    # This is testing that the unnamed args specified
    # in the block definition are used as names for the
    # unnamed arguments passed to the block.

    source = """
    ::defblock:alias, blocktype, attr1, attr2

    [alias, value1, value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"attr1": "value1", "attr2": "value2"},
        ),
    ]


def test_block_definitions_unnamed_and_named_args():
    # This is testing that any named are specifid in
    # the block definition is included in the block
    # kwargs if not redefined.

    source = """
    ::defblock:alias, blocktype, attr1, attr2, attr3=value3

    [alias, value1, value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={"attr1": "value1", "attr2": "value2", "attr3": "value3"},
        ),
    ]


def test_block_definitions_no_values_for_unnamed_args():
    # This is testing that is there are unnamed args in
    # the block definiton they should be given values
    # when the block is created.

    source = """
    ::defblock:alias, blocktype, attr1, attr2, attr3=value3

    [alias]
    ----
    ----
    """

    with pytest.raises(ValueError):
        runner(source)


def test_block_definitions_values_without_unnamed_args():
    # This is testing that values passed to the
    # block at creation time should match the parameters
    # specified in the block definition.

    source = """
    ::defblock:alias, blocktype, attr3=value3

    [alias, value1, value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="blocktype",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=["value1", "value2"],
            kwargs={"attr3": "value3"},
        ),
    ]



import pytest
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.inline import (
    ClassNode,
    ImageNode,
    LinkNode,
    ListItemNode,
    MacroNode,
    SentenceNode,
    StyleNode,
    TextNode,
    VerbatimNode,
)
from mau.nodes.references import ReferenceNode
from mau.visitors.base_visitor import BaseVisitor, VisitorError


def test_no_node():
    visitor = BaseVisitor()

    result = visitor.visit(None)

    assert result == {}


def test_unknown_node():
    visitor = BaseVisitor()

    node = TextNode("Just some text.")
    node.node_type = "unknown"

    with pytest.raises(VisitorError):
        visitor.visit(node)


def test_text_node():
    visitor = BaseVisitor()

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "text",
            "value": "Just some text.",
        }
    }


def test_verbatim_node():
    visitor = BaseVisitor()

    node = VerbatimNode("Just some text.")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "verbatim",
            "value": "Just some text.",
        }
    }


def test_sentence_node():
    visitor = BaseVisitor()

    node = SentenceNode(
        [
            TextNode("Just some text. "),
            TextNode("More text."),
        ]
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "sentence",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text. ",
                    }
                },
                {
                    "data": {
                        "type": "text",
                        "value": "More text.",
                    }
                },
            ],
        }
    }


def test_style_node_star():
    visitor = BaseVisitor()

    node = StyleNode("star", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": {
                "data": {
                    "type": "sentence",
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Just some text.",
                            }
                        },
                    ],
                }
            },
            "value": "star",
        }
    }


def test_style_node_underscore():
    visitor = BaseVisitor()

    node = StyleNode("underscore", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": {
                "data": {
                    "type": "sentence",
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Just some text.",
                            }
                        },
                    ],
                }
            },
            "value": "underscore",
        }
    }


def test_style_node_tilde():
    visitor = BaseVisitor()

    node = StyleNode("tilde", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": {
                "data": {
                    "type": "sentence",
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Just some text.",
                            }
                        },
                    ],
                }
            },
            "value": "tilde",
        }
    }


def test_style_node_caret():
    visitor = BaseVisitor()

    node = StyleNode("caret", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": {
                "data": {
                    "type": "sentence",
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Just some text.",
                            }
                        },
                    ],
                }
            },
            "value": "caret",
        }
    }


def test_macro_node():
    visitor = BaseVisitor()

    node = MacroNode("somename", ["arg1", "arg2"], {"key1": "value1"})

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro",
            "name": "somename",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
        }
    }


def test_footnote_node():
    visitor = BaseVisitor()

    node = FootnoteNode(
        content=[TextNode("Just some text.")],
        number="5",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "footnote",
            "number": "5",
            "reference_anchor": "someanchor",
            "content_anchor": "someanchor-def",
            "content": [
                {"data": {"type": "text", "value": "Just some text."}},
            ],
        }
    }


def test_reference_node():
    visitor = BaseVisitor()

    node = ReferenceNode(
        content_type="somecontent",
        name="somename",
        category="somecategory",
        content=[TextNode("Just some text.")],
        number="5",
        title=SentenceNode([TextNode("Some title")]),
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "reference",
            "category": "somecategory",
            "content_type": "somecontent",
            "name": "somename",
            "number": "5",
            "reference_anchor": "someanchor",
            "content_anchor": "someanchor-def",
            "content": [
                {
                    "data": {"type": "text", "value": "Just some text."},
                }
            ],
            "title": {
                "data": {
                    "type": "sentence",
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Some title",
                            },
                        }
                    ],
                }
            },
        }
    }


def test_class_node():
    visitor = BaseVisitor()

    node = ClassNode(["class1", "class2"], TextNode("Just some text."))

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "class",
            "content": {
                "data": {
                    "type": "text",
                    "value": "Just some text.",
                }
            },
            "classes": ["class1", "class2"],
        }
    }


def test_link_node():
    visitor = BaseVisitor()

    node = LinkNode(target="sometarget", text="sometext")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "link",
            "text": "sometext",
            "target": "sometarget",
        }
    }


def test_image_node():
    visitor = BaseVisitor()

    node = ImageNode(uri="someuri", alt_text="sometext", width="100", height="400")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "image",
            "uri": "someuri",
            "alt_text": "sometext",
            "width": "100",
            "height": "400",
        }
    }


def test_list_item_node():
    visitor = BaseVisitor()

    node = ListItemNode("4", TextNode("Just some text."))

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "list_item",
            "level": 4,
            "content": {
                "data": {
                    "type": "text",
                    "value": "Just some text.",
                }
            },
        }
    }
